// ======================================================================
//
// MenuBitmap.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "MenuBitmap.h"

#include "Bitmap.h"
#include "Iff.h"
#include "Mouse.h"

// ======================================================================

const Tag TAG_BITS = TAG(B,I,T,S);
const Tag TAG_RNGE = TAG(R,N,G,E);
const Tag TAG_BEGF = TAG(B,E,G,F);
const Tag TAG_ABIA = TAG(A,B,I,A); // animating bitmap int animation selector
const Tag TAG_B3IS = TAG(B,3,I,S); // animating bitmap shared integer
const Tag TAG_AB3E = TAG(A,B,3,E); // animating bitmap animation sequence entry
const int MENU_INT_ANIM_STOP         = -1;
const int MENU_INT_ANIM_KEEP_LOOPING = -2;

// ======================================================================

MenuBitmap::MenuBitmap(MenuElement *newParent, int maxWidth, int maxHeight, Iff *iff)
: MenuElement(newParent, maxWidth, maxHeight, NULL)
{
	bitmap = NULL;
	bitmapFrame = 0;
	ownsBitmap = true;
	isMouse = false;

	load(iff);
}

// ----------------------------------------------------------------------

MenuBitmap::~MenuBitmap(void)
{
	if (ownsBitmap)
		delete bitmap;
}

// ----------------------------------------------------------------------

void MenuBitmap::load(Iff *iff)
{
	if (iff)
	{
		MenuElement::load(iff);

		if (iff->seekChunk(TAG_MOUS, Iff::FlagOptional))
		{
			isMouse = true;
			iff->exitChunk();
		}

		if (iff->seekForm(TAG_BITM, Iff::FlagOptional))
		{
			bitmap = NEW Bitmap(*iff);
			iff->exitForm();
		}
		else
			if (iff->seekChunk(TAG_BITS, Iff::FlagOptional))
			{
				char *sharedName;

				sharedName = iff->read_char(Iff::ReadRest);
				iff->exitChunk();

				bitmap = findSharedBitmap(sharedName);
				DEBUG_FATAL(!bitmap, ("MENU_BITMAP cannot find shared bitmap '%s'", sharedName));

				ownsBitmap = false;
				DEBUG_FATAL(!bitmap, ("MenuBitmap::load shared bitmap '%s' not found", sharedName));

				delete [] sharedName;
			}
			else
			{
				iff->seekChunk(TAG_BITN);

				// -qq- load a named bitmap
				DEBUG_FATAL(true, ("MenuBitmap::load BITN not yet supported"));

				iff->exitChunk();
			}

		if (iff->seekChunk(TAG_FRAM, Iff::FlagOptional))
		{
			int f;

			f = iff->read_int32();

			if (f < 0)
				f = getCurrentBitmap()->frames() + f;

			setCurrentFrame(f);
			iff->exitChunk();
		}
	}
}

// ----------------------------------------------------------------------

Bitmap *MenuBitmap::getCurrentBitmap(void) const
{
	return bitmap;
}

// ----------------------------------------------------------------------

void MenuBitmap::setCurrentFrame(int newBitmapFrame)
{
	bitmapFrame = newBitmapFrame;
}

// ----------------------------------------------------------------------

int MenuBitmap::getCurrentFrame(void) const
{
	return bitmapFrame;
}

// ----------------------------------------------------------------------

IoResult MenuBitmap::event(IoEventStruct *event)
{
	if (isMouse && event->type == IoEventMouseSetCursor && menuRect.isInInclusive(event->mouseX, event->mouseY))
	{
		Mouse::setBitmap(getCurrentBitmap(), getCurrentFrame());
		return IoResultBlock;
	}

	return MenuElement::event(event);
}

// ----------------------------------------------------------------------

void MenuBitmap::menuDraw(int x, int y) const
{
	Bitmap *bm;

	if (!isMouse && (bm = getCurrentBitmap()) != NULL)
		bm->draw(x+menuRect.x0, y+menuRect.y0, getCurrentFrame());

	MenuElement::menuDraw(x, y);
}

// ======================================================================

MenuAnimatingBitmap::MenuAnimatingBitmap(MenuElement *parent, int maxWidth, int maxHeight, Iff *iff)
: MenuBitmap(parent, maxWidth, maxHeight, NULL)
{
	notAnimatingFrame = 0;
	animatingStyle    = forwardLooping;
	currentState      = forwardLooping;
	beginningEvent    = 0;
	endingEvent       = 0;
	step              = 1;
	startFrame        = 0;
	endFrame          = 0;

	load(iff);
}

// ----------------------------------------------------------------------

MenuAnimatingBitmap::~MenuAnimatingBitmap(void)
{
}

// ----------------------------------------------------------------------

void MenuAnimatingBitmap::load(Iff *iff)
{
	int sf;

	if (iff)
	{
		MenuBitmap::load(iff);

		// this variable keeps track of which frame to display when it's NOT animating
		notAnimatingFrame = getCurrentFrame();

		if (iff->seekChunk(TAG_SEC, Iff::FlagOptional))
		{
			real seconds;

			seconds = iff->read_float();
			timer.setTime(seconds);
			iff->exitChunk();
		}
		else
		{
			real hz;

			(void)iff->seekChunk(TAG_HZ);
			hz = iff->read_float();
			timer.setHz(hz);
			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_STYL, Iff::FlagOptional))
		{
			currentState = animatingStyle = static_cast<AnimatingStyle>(iff->read_int32());

			switch (animatingStyle)
			{
				case stop:
					step = 0;
					break;

				case forwardStop:
				case forwardLooping:
				case pingPong:
				case pingPongOnce:
					step = 1;
					break;

				case backwardStop:
				case backwardLooping:
					if (getCurrentBitmap())
						setCurrentFrame(getCurrentBitmap()->frames()-1);
					step = -1;
					break;

				default:
					DEBUG_FATAL(true, ("MenuAnimBitmap::load bad style %d", animatingStyle));
			}

			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_BEGE, Iff::FlagOptional))
		{
			// we DON'T want the animation to start until this happens
			currentState = stop;
			beginningEvent = iff->read_int32();
			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_ENDE, Iff::FlagOptional))
		{
			endingEvent = iff->read_int32();
			iff->exitChunk();
		}

		endFrame = getCurrentBitmap()->frames() - 1;
		if (iff->seekChunk(TAG_RNGE, Iff::FlagOptional))
		{
			startFrame = iff->read_int32();
			endFrame   = iff->read_int32();
			if (startFrame < 0 || endFrame > getCurrentBitmap()->frames() - 1)
				Fatal("MenuAnimatingBitmap::load invalid range: (%d to %d), bitmap only contains (0 to %d)", startFrame, endFrame, getCurrentBitmap()->frames() - 1);
			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_BEGF, Iff::FlagOptional))
		{
			sf = iff->read_int32();
			if (sf < startFrame || sf >= endFrame)
				Fatal("MenuAnimatingBitmap::load invalid start frame: '%d' valid range:(%d to %d)", sf, startFrame, endFrame);
			setCurrentFrame(sf);
			iff->exitChunk();
		}
	}
}

// ----------------------------------------------------------------------

void MenuAnimatingBitmap::menuEvent(int id, MenuElement *whom)
{
	UNREF(whom);

	if (beginningEvent != 0 && id == beginningEvent)
	{
		// reset timer here as well, or is it not needed??? -ck
		currentState = animatingStyle;
		setCurrentFrame(startFrame);
		step = 1;
	}

	if (endingEvent != 0 && id == endingEvent)
	{
		setCurrentFrame(notAnimatingFrame);
		currentState = stop;
	}

	MenuElement::menuEvent(id, whom);
}

// ----------------------------------------------------------------------

IoResult MenuAnimatingBitmap::update(void)
{
	real    t;
	int     f;
	Bitmap *bm;

	if (currentState != stop)
	{
		bm = getCurrentBitmap();
		f  = getCurrentFrame();
		t  = Clock::frameTime();

		DEBUG_FATAL(!bm, ("MenuAnimBitmap::update no bitmap"));

		while (timer.updateSub(t))
		{
			t = CONST_REAL(0);

			f += step;
			
			if (f < startFrame || f >= endFrame)
			{
				switch (currentState)
				{
					case forwardStop:
						f--;
						currentState = stop;
						step = 0;
						break;

					case backwardStop:
						f++;
						currentState = stop;
						step = 0;
						break;
						
					case forwardLooping:
						f = startFrame;
						break;

					case backwardLooping:
						f = endFrame;
						break;

					case pingPong:
						if (step < startFrame)
							f++;
						else
							f--;
						step = (-step);
						break;

					case pingPongOnce:
						// ping pong once, then stop
						if (step < startFrame)
						{
							f++;
							currentState = stop;
							step = 0;
						}
						else
						{
							f--;
							step = (-step);
						}
						break;

					default:
						DEBUG_FATAL(true, ("MenuAnimatingBitmap::update bad style %d", currentState));
				}
			}
		}

		setCurrentFrame(f);
	}

	return MenuBitmap::update();
}

// ======================================================================

MenuAnimatingBitmap2::MenuAnimatingBitmap2(MenuElement *parent, int maxWidth, int maxHeight, Iff *iff)
: MenuBitmap(parent, maxWidth, maxHeight, NULL)
{
	integerAnimation = NULL;
	load(iff);
}

// ----------------------------------------------------------------------

MenuAnimatingBitmap2::~MenuAnimatingBitmap2(void)
{
	integerAnimation = NULL;
}

// ----------------------------------------------------------------------

void MenuAnimatingBitmap2::load(Iff *iff)
{
	char *name;

	if (iff)
	{
		MenuBitmap::load(iff);

		iff->seekChunk(TAG_ABIA);
		name = iff->read_char(Iff::ReadRest);
		if ((integerAnimation = findSharedIntegerAnimation(name)) == NULL)
			Fatal("MenuAnimatingBitmapSelector::load cannot find menuintanimatingselector '%s'", name);
		iff->exitChunk(); // 'ABIA'

		delete []name;
		name = NULL;
	}
}

// ----------------------------------------------------------------------

int MenuAnimatingBitmap2::getCurrentFrame(void) const
{
	return integerAnimation->getFrame();
}

// ======================================================================


MenuAnimatingBitmap3::MenuAnimatingBitmap3(MenuElement *parent, int maxWidth, int maxHeight, Iff *iff)
: MenuBitmap(parent, maxWidth, maxHeight, NULL)
{
	sharedInteger     = NULL;
	animationSequence = NULL;
	numSequences      = 0;
	currentSequence   = 0;
	load(iff);
}

// ----------------------------------------------------------------------

MenuAnimatingBitmap3::~MenuAnimatingBitmap3(void)
{
	int i;

	for (i = 0; i < numSequences; i++)
	{
		delete []animationSequence[i];
		animationSequence[i] = NULL;
	}
	delete []animationSequence;
	animationSequence = NULL;
}

// ----------------------------------------------------------------------

void MenuAnimatingBitmap3::load(Iff *iff)
{
	real seconds;
	real hz;
	char *name;
	int  suggestedMinimumEntries = 1;
	int  i;

	if (iff)
	{
		MenuBitmap::load(iff);

		if (iff->seekChunk(TAG_SEC, Iff::FlagOptional))
		{
			seconds = iff->read_float();
			timer.setTime(seconds);
			iff->exitChunk();
		}
		else
		{
			(void)iff->seekChunk(TAG_HZ);
			hz = iff->read_float();
			timer.setHz(hz);
			iff->exitChunk();
		}

		// extract the shared integer name
		iff->seekChunk(TAG_B3IS);
		name = iff->read_char(Iff::ReadRest);
		sharedInteger = findSharedInteger(name);
		DEBUG_FATAL(!sharedInteger, ("MENU_ANIM_BITMAP3 cannot find shared int '%s'", name));

		iff->exitChunk(); // 'B3IS'
		delete []name;
		name = NULL;

		// count the # of animation sequence
		if ((numSequences = iff->blockCount(Iff::FlagChunk, TAG_AB3E)) <= suggestedMinimumEntries)
			Fatal("MenuTextVar. load must contain at least %d entries to be effective.", suggestedMinimumEntries);
		animationSequence = NEW int *[numSequences];
		for (i = 0; iff->seekChunk(TAG_AB3E, Iff::FlagOptional); i++)
		{
			animationSequence[i] = iff->read_int32(Iff::ReadRest);
			iff->exitChunk(); // 'AB3E'
		}

		// set up our initial current frame
		currentSequence = sharedInteger->getValue();
		currentFrame    = 0;
		setCurrentFrame(animationSequence[currentSequence][currentFrame]);
	}
}

// ----------------------------------------------------------------------

IoResult MenuAnimatingBitmap3::update(void)
{
	real    t;

	t = Clock::frameTime();
	while (timer.updateSub(t))
	{
		t = CONST_REAL(0);

		currentFrame++;
		if (currentSequence != sharedInteger->getValue())
		{
			currentSequence = sharedInteger->getValue();
			currentFrame = 0;
		}

		switch (animationSequence[currentSequence][currentFrame])
		{
			case MENU_INT_ANIM_STOP:         // stop
				currentFrame--;
				break;
			case MENU_INT_ANIM_KEEP_LOOPING: // keep looping
				currentFrame = 0;
				break;
		}

		// set up our current frame
		setCurrentFrame(animationSequence[sharedInteger->getValue()][currentFrame]);
	}

	return MenuBitmap::update();
}

// ======================================================================
